/*
 *
 * stars3d_space.c - v0.5 - 2005-01-10 - MinGW32 support.
 *                   v0.2 - 2002-02-14
 *                   v0.1 - 2000-10-12
 *
 * Clemens Fuchslocher <clemens@allesdurcheinander.de>
 *
 */

/*
 *
 * SDL - Simple DirectMedia Layer <http://www.libsdl.org/>
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>

#include <SDL.h>

#ifdef __MINGW32__
#define SDL_main main
#endif

#define PERIOD 	197 /* T=2*pi/w, w=1/10*pi */
#define STARS	800
#define FPS     1000/60

typedef struct Star_tag
{
	float x, y, z;
} Star;

void new_star (Star *star, SDL_Surface *surface);
void put_pixel (Uint16 x, Uint16 y, Uint32 color, SDL_Surface *surface);

int main (void)
{
	SDL_Surface *screen;

	Uint32 i, frames = 0, t = 0;

	Star stars[STARS];
	float x, y, z;
	float x_t, y_t, z_t;
	float x_new, y_new;
	Sint32 color;

	Uint32 this_tick_count;
	Uint32 last_tick_count;

	float sin_table[PERIOD];
	float cos_table[PERIOD];

	time_t runtime = time (NULL);
	srand (time (NULL));

#ifdef __MINGW32__
	FILE *fps;
#endif

	for (i = 0; i < PERIOD; i++)
	{
		sin_table[i] = sin (i / (10 * 3.14));
		cos_table[i] = cos (i / (10 * 3.14));
	}

	if (SDL_Init (SDL_INIT_VIDEO) < 0)
	{
		fprintf (stderr, ".oO ( Oops : %s )\n", SDL_GetError ());
		exit (EXIT_FAILURE);
	}
	atexit (SDL_Quit);

	for (i = 0; i < SDL_NUMEVENTS; i++)
	{
		if ((i != SDL_KEYDOWN) && (i != SDL_QUIT))
		{
			SDL_EventState(i, SDL_IGNORE);
		}
	}

	screen = SDL_SetVideoMode (800, 600, 32, SDL_HWSURFACE | SDL_FULLSCREEN | SDL_DOUBLEBUF);
	if (screen == NULL)
	{
		fprintf (stderr, ".oO ( Oops: %s )\n", SDL_GetError ());
		exit (EXIT_FAILURE);
	}

	SDL_WM_SetCaption ("stars3d space", "stars3d space");
	SDL_ShowCursor (0);

	for (i = 0; i < STARS; i++)
	{
		new_star (&stars[i], screen);
	}

	last_tick_count = SDL_GetTicks ();
	while (SDL_PollEvent (NULL) == 0)
	{
		frames++;
		if (SDL_MUSTLOCK (screen))
		{
			if (SDL_LockSurface (screen) < 0)
			{
				fprintf (stderr, ".oO ( Oops: %s )\n", SDL_GetError ());
				exit (EXIT_FAILURE);
			}
		}

		t = (t + 1) % PERIOD;
		for (i = 0; i < STARS; i++)
		{
			x = stars[i].x;
			y = stars[i].y;
			z = stars[i].z;

			/* Mathematische Formelsammlung, Lothar Papula, 5. Auflage */
			/* 9.1.3.3 Drehung eines kartesischen Koordinatensystems   */

			/* rotation um y achse */
			x_t = (x * cos_table[t] + z * sin_table[t]);
			z_t = (x * -sin_table[t] + z * cos_table[t]);

			x = x_t;
			z = z_t;

			/* rotation um z achse */
			x_t = (x * cos_table[t] + y * -sin_table[t]);
			y_t = (x * sin_table[t] + y * cos_table[t]);

			x = x_t;
			y = y_t;

			/* rotation um x achse */
			y_t = (y * cos_table[t] + z * -sin_table[t]);
			z_t = (y * sin_table[t] + z * cos_table[t]);

			y = y_t;
			z = z_t;

			color = z + 140;
			if (color > 255) color = 255;
			else if (color < 0) color = 0;

			color = ((int) color << screen->format->Rshift) |
				((int) color << screen->format->Gshift) |
				((int) color << screen->format->Bshift);

			/* 3d -> 2d projektion (mc extra 3/95) */
			x_new = 250 * (x / (z / 2 - 300)) + screen->w/2;
			y_new = 250 * (y / (z / 2 - 300)) + screen->h/2;

			put_pixel (x_new,   y_new,   color, screen);
			put_pixel (x_new-1, y_new,   color, screen);
			put_pixel (x_new,   y_new-1, color, screen);
			put_pixel (x_new+1, y_new,   color, screen);
			put_pixel (x_new,   y_new+1, color, screen);
		}

		if (SDL_MUSTLOCK (screen))
		{
			SDL_UnlockSurface (screen);
		}

		this_tick_count = SDL_GetTicks ();
		if ((this_tick_count - last_tick_count) < FPS)
		{
			SDL_Delay (FPS - (this_tick_count - last_tick_count));
		}
		last_tick_count = this_tick_count;

		SDL_Flip (screen);
		SDL_FillRect (screen, NULL, 0);
	}

#ifdef __MINGW32__
	fps = fopen ("fps.txt", "w");
	if (fps != NULL)
	{
		fprintf (fps, "fps    : %ld\n", frames / (time (NULL) - runtime));
		fprintf (fps, "runtime: %ld\n", time (NULL) - runtime);
		fclose (fps);
	}
#else
	printf ("fps    : %ld\n", frames / (time (NULL) - runtime));
	printf ("runtime: %ld\n", time (NULL) - runtime);
#endif

	SDL_ShowCursor (1);
	SDL_Quit ();

	return EXIT_SUCCESS;
}

void new_star (Star *star, SDL_Surface *surface)
{
	star->x = (rand () % 320) - 160;
	star->y = (rand () % 320) - 160;
	star->z = (rand () % 320) - 160;
}

void put_pixel (Uint16 x, Uint16 y, Uint32 color, SDL_Surface *surface)
{
	if (x < surface->w)
	{
		if (y < surface->h)
		{
			*((Uint32 *) surface->pixels + (y * surface->pitch/4 + x)) = color;
		}
	}
}

