/*
 *
 * stars3d_transformation.c - v0.5 - 2005-01-10 - MinGW32 support.
 *                            v0.3 - 2002-02-14
 *                            v0.2 - 2000-10-17
 *
 * Clemens Fuchslocher <clemens@allesdurcheinander.de>
 *
 */

/*
 *
 * SDL - Simple DirectMedia Layer <http://www.libsdl.org/>
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>

#include <SDL.h>

#ifdef __MINGW32__
#define SDL_main main
#endif

#define SPEED_T 40.0
#define PERIOD  197 /* T=2*pi/w, w=1/10*pi */
#define STARS	800
#define NEXT	500
#define FPS	1000/60

typedef struct Star_tag
{
	float x, y, z;
} Star;

Star orb[STARS];
Star pipe[STARS];
Star cube[STARS];
Star cone[STARS];
Star plane[STARS];
Star helix[STARS];
Star stars[STARS];
Star square[STARS];
Star cylinder[STARS];

Star *object[9];
Uint32 which_object = 0, next_object = NEXT;

float sin_table[PERIOD];
float cos_table[PERIOD];

void init (void);
void put_pixel (Uint16 x, Uint16 y, Uint32 color, SDL_Surface *surface);

int main (void)
{
	SDL_Surface *screen;

	Sint32 i, frames = 0, t = 0;

	float x, y, z;
	float x_t, y_t, z_t;
	float x_delta, y_delta, z_delta;
	float x_new, y_new;
	Sint32 color;

	Uint32 this_tick_count;
	Uint32 last_tick_count;

	time_t runtime = time (NULL);
	srand (time (NULL));

#ifdef __MINGW32__
	FILE *fps;
#endif

	init ();

	if (SDL_Init (SDL_INIT_VIDEO) < 0)
	{
		fprintf (stderr, ".oO ( Oops : %s )\n", SDL_GetError ());
		exit (EXIT_FAILURE);
	}
	atexit (SDL_Quit);

	for (i = 0; i < SDL_NUMEVENTS; i++)
	{
		if ((i != SDL_KEYDOWN) && (i != SDL_QUIT))
		{
			SDL_EventState(i, SDL_IGNORE);
		}
	}

	screen = SDL_SetVideoMode (800, 600, 32, SDL_HWSURFACE | SDL_FULLSCREEN | SDL_DOUBLEBUF);
	if (screen == NULL)
	{
		fprintf (stderr, ".oO ( Oops: %s )\n", SDL_GetError ());
		exit (EXIT_FAILURE);
	}

	SDL_WM_SetCaption ("stars3d space transformation", "stars3d space transformation");
	SDL_ShowCursor (0);

	last_tick_count = SDL_GetTicks ();
	while (SDL_PollEvent (NULL) == 0)
	{
		frames++;
		if (SDL_MUSTLOCK (screen))
		{
			if (SDL_LockSurface (screen) < 0)
			{
				fprintf (stderr, ".oO ( Oops: %s )\n", SDL_GetError ());
				exit (EXIT_FAILURE);
			}
		}

		t = (t + 1) % PERIOD;
		for (i = 0; i < STARS; i++)
		{
			/* Transformation durch 'Koordinatenannaeherung' */
			x_delta = (object[which_object][i].x - stars[i].x) / SPEED_T;
			y_delta = (object[which_object][i].y - stars[i].y) / SPEED_T;
			z_delta = (object[which_object][i].z - stars[i].z) / SPEED_T;

			stars[i].x += x_delta;
			stars[i].y += y_delta;
			stars[i].z += z_delta;

			if (!x_delta) stars[i].x = object[which_object][i].x;
			if (!y_delta) stars[i].y = object[which_object][i].y;
			if (!z_delta) stars[i].z = object[which_object][i].z;

			if (!next_object)
			{
				if (object[++which_object] == NULL)
				{
					which_object = 0;
				}
				next_object = NEXT;
			}

			x = stars[i].x / 500;
			y = stars[i].y / 500;
			z = stars[i].z + 500;

			/* Mathematische Formelsammlung, Lothar Papula, 5. Auflage */
			/* 9.1.3.3 Drehung eines kartesischen Koordinatensystems   */

			/* rotation um y achse */
			x_t = (x * cos_table[t] + z * sin_table[t]);
			z_t = (x * -sin_table[t] + z * cos_table[t]);

			x = x_t;
			z = z_t;

			/* rotation um z achse */
			x_t = (x * cos_table[t] + y * -sin_table[t]);
			y_t = (x * sin_table[t] + y * cos_table[t]);

			x = x_t;
			y = y_t;

			/* rotation um x achse */
			y_t = (y * cos_table[t] + z * -sin_table[t]);
			z_t = (y * sin_table[t] + z * cos_table[t]);

			y = y_t;
			z = z_t;

			color = z + 140;
			if (color > 255) color = 255;
			else if (color < 0) color = 0;

			color = ((int) color << screen->format->Rshift) |
				((int) color << screen->format->Gshift) |
				((int) color << screen->format->Bshift);

			/* 3d -> 2d projektion (mc extra 3/95) */
			x_new = 300 * (x / (z / 2 - 300)) + screen->w/2;
			y_new = 300 * (y / (z / 2 - 300)) + screen->h/2;

			put_pixel (x_new,   y_new,   color, screen);
			put_pixel (x_new-1, y_new,   color, screen);
			put_pixel (x_new,   y_new-1, color, screen);
			put_pixel (x_new+1, y_new,   color, screen);
			put_pixel (x_new,   y_new+1, color, screen);
		}
		next_object--;

		if (SDL_MUSTLOCK (screen))
		{
			SDL_UnlockSurface (screen);
		}

	   	this_tick_count = SDL_GetTicks ();
		if ((this_tick_count - last_tick_count) < FPS)
		{
			SDL_Delay (FPS - (this_tick_count - last_tick_count));
		}
		last_tick_count = this_tick_count;

		SDL_Flip (screen);
		SDL_FillRect (screen, NULL, 0);
	}

#ifdef __MINGW32__
	fps = fopen ("fps.txt", "w");
	if (fps != NULL)
	{
		fprintf (fps, "fps    : %ld\n", frames / (time (NULL) - runtime));
		fprintf (fps, "runtime: %ld\n", time (NULL) - runtime);
		fclose (fps);
	}
#else
	printf ("fps    : %ld\n", frames / (time (NULL) - runtime));
	printf ("runtime: %ld\n", time (NULL) - runtime);
#endif

	SDL_ShowCursor (1);
	SDL_Quit ();

	return EXIT_SUCCESS;
}

void put_pixel (Uint16 x, Uint16 y, Uint32 color, SDL_Surface *surface)
{
	if (x < surface->w)
	{
		if (y < surface->h)
		{
			*((Uint32 *) surface->pixels + (y * surface->pitch/4 + x)) = color;
		}
	}
}

void init (void)
{
	int i, x, y, z;

	for (i = 0; i < PERIOD; i++)
	{
		sin_table[i] = sin (i / (10 * 3.14));
		cos_table[i] = cos (i / (10 * 3.14));
	}

	object[0] = cube;
	object[1] = helix;
	object[2] = cone;
	object[3] = plane;
	object[4] = pipe;
	object[5] = cylinder;
	object[6] = square;
	object[7] = orb;
	object[8] = NULL;

	for (i = 0; i < STARS; i++)
	{
		cube[i].x = stars[i].x = (100 * sin (i / 31.4)) * 500;
		cube[i].y = stars[i].y = (100 * cos (stars[i].x / 31.4)) * 500;
		cube[i].z = stars[i].z = (100 * sin (stars[i].y / 31.4)) - 500;

		helix[i].x = (i/4 - 100) * 500;
		helix[i].y = ((40 * cos (i / 31.4))) * 500;
		helix[i].z = ((40 * sin (i / 31.4))) - 500;

		plane[i].x = ((rand () % 320) - 160) * 500;
		plane[i].y = 500;
		plane[i].z = ((rand () % 320) - 160) - 500;

		pipe[i].x = (i/5 * sin (i / 31.4)) * 500;
		pipe[i].y = (i/5) * 500;
		pipe[i].z = (i/5 * cos (i / 31.4)) - 500;
	}

	i = 0;
	for (z = 0; z < 8; z++)
	{
		for (y = 0; y < 10; y++)
		{
			for (x = 0; x < 10; x++)
			{
				square[i].x = (x * 25 - 100) * 500;
				square[i].y = (y * 25 - 100) * 500;
				square[i].z = (z * 30 - 80) - 500;
				i++;
			}
		}
	}

	for (y = 0; y < 16; y++)
	{
		for (x = 0; x < 50; x++)
		{
			z = (y * 50) + x;
			cylinder[z].x = (70 * cos (x)) * 500;
			cylinder[z].y = (70 * sin (x)) * 500;
			cylinder[z].z = y * 20 - 160 - 500;

			cone[z].x = ((5 * y) * cos (x)) * 500;
			cone[z].y = ((5 * y) * sin (x)) * 500;
			cone[z].z = (y * 20) - 160 - 500;

			orb[z].x = (100 * sin (y/(16.0/3.14)) * cos (x)) * 500;
			orb[z].y = (100 * sin (y/(16.0/3.14)) * sin (x)) * 500;
			orb[z].z = y * 16 - 128 - 500;
		}
	}
}

